# Python 2.7 Directives
from __future__ import absolute_import

# Python 2.7 Standard Library
import pandoc.doctest
import doctest
import unittest
import sys
import tempfile

# Third-Party Libraries
import pkg_resources


path = pkg_resources.resource_filename("pandoc", "tests.md")
suite = doctest.DocFileSuite(path, module_relative=False)


def run(**kwargs):
    return doctest.testfile(path, module_relative=False, **kwargs)



# Python 2.7 Standard Library
from __future__ import absolute_import
from __future__ import print_function
import json

# Third-Party Libraries
import pkg_resources
import ply.lex as lex
import ply.yacc as yacc


# Pandoc-Types Version Mapping and Type Info
# ------------------------------------------------------------------------------
_json_data = pkg_resources.resource_string("pandoc", "pandoc-types.js")
if not isinstance(_json_data, str):  # resource loaded as bytes in Python 3
    _json_data = _json_data.decode("utf-8")
_data = json.loads(_json_data)
version_mapping = _data["version_mapping"]
definitions = _data["definitions"]


# Pandoc-Types Version Resolver
# ------------------------------------------------------------------------------
def version_key(string):
    return [int(s) for s in string.split(".")]


def match(spec, version):
    if len(spec) == 0 or (len(spec) >= 1 and isinstance(spec[0], list)):
        return all(match(s, version) for s in spec)
    elif spec[0] == "==":
        if "*" in spec[1]:
            vk_low = version_key(spec[1][:-2])
            vk_high = vk_low.copy()
            vk_high[-1] += 1
            return match(
                [
                    [">=", ".".join(p for p in vk_low)],
                    ["<", ".".join(p for p in vk_high)],
                ],
                version,
            )
        else:
            return spec[1] == version
    elif spec[0] == ">=":
        return version_key(version) >= version_key(spec[1])
    elif spec[0] == "<":
        return version_key(version) < version_key(spec[1])
    else:
        raise ValueError("invalid version spec {0}".format(spec))


def resolve(version):
    try:
        pt_spec = version_mapping[version]
    except KeyError:
        error = "pandoc {0} is not registered"
        raise ValueError(error.format(version))
    pandoc_types_versions = sorted(definitions.keys(), key=version_key)

    matches = []
    for pt_version in pandoc_types_versions:
        if match(pt_spec, pt_version):
            matches.append(pt_version)
    return matches


# Lexer
# ------------------------------------------------------------------------------
tokens = [
    "CONID",
    "VARID",
    "COMMA",
    "BAR",
    "EQUAL",
    "DCOLON",
    "LPAREN",
    "RPAREN",
    "LBRACKET",
    "RBRACKET",
    "LBRACE",
    "RBRACE",
]
keywords = {
    "data": "DATA",
    "type": "TYPE",
    "newtype": "NEWTYPE",
    "Map": "MAP",
}
tokens = tokens + list(keywords.values())


def t_CONID(t):
    r"[A-Z][a-zA-Z_0-9`']*"
    t.type = keywords.get(t.value, "CONID")
    return t


def t_VARID(t):
    r"[a-z][a-zA-Z_0-9`']*"
    t.type = keywords.get(t.value, "VARID")
    return t


t_COMMA = r"\,"
t_BAR = r"\|"
t_EQUAL = r"\="
t_DCOLON = r"\:\:"
t_LPAREN = r"\("
t_RPAREN = r"\)"
t_LBRACKET = r"\["
t_RBRACKET = r"\]"
t_LBRACE = r"\{"
t_RBRACE = r"\}"

t_ignore = " \t\n"


def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)


lexer = lex.lex()

# Parser
# ------------------------------------------------------------------------------
def p_typedecl(p):
    """typedecl : typetypedecl
                | datatypedecl
                | newtypedecl"""
    p[0] = p[1]


def p_typetypedecl(p):
    "typetypedecl : TYPE CONID EQUAL type"
    p[0] = [p[1], [p[2], p[4]]]


def p_type_paren(p):
    "type : LPAREN type RPAREN"
    p[0] = p[2]


def p_type_conid(p):
    "type : CONID"
    p[0] = p[1]


def p_type_list(p):
    "type : LBRACKET type RBRACKET"
    p[0] = ["list", [p[2]]]


def p_comma_separated_types_2(p):
    "comma_separated_types : type COMMA type"
    p[0] = [p[1]] + [p[3]]


def p_comma_separated_types_more(p):
    "comma_separated_types : type COMMA comma_separated_types"
    p[0] = [p[1]] + p[3]


def p_type_tuple(p):
    "type : LPAREN comma_separated_types RPAREN"
    p[0] = ["tuple", p[2]]


def p_type_map(p):
    "type : MAP type type"
    p[0] = ["map", [p[2], p[3]]]


def p_assignment(p):
    """
    assignment : VARID DCOLON type
    """
    p[0] = [p[1], p[3]]


def p_assignments(p):
    """
    assignments : assignment
                | assignment COMMA assignments
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = [p[1]] + p[3]


def p_record(p):
    """type_record : LBRACE RBRACE
                   | LBRACE assignments RBRACE
    """
    if len(p) == 3:
        p[0] = ["map", []]
    else:
        p[0] = ["map", p[2]]


def p_types(p):
    """types : type
             | type types
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = [p[1]] + p[2]


def p_constructor(p):
    """constructor : CONID types
                   | CONID
                   | CONID type_record
    """
    if len(p) == 3 and p[2][0] == "map":
        p[0] = [p[1], p[2]]
    else:
        if len(p) == 2:
            p[0] = [p[1], ["list", []]]
        else:
            p[0] = [p[1], ["list", p[2]]]


def p_constructors(p):
    """constructors : constructor
                    | constructor BAR constructors
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = [p[1]] + p[3]


def p_datatypedecl(p):
    "datatypedecl : DATA CONID EQUAL constructors"
    p[0] = [p[1], [p[2], p[4]]]


def p_newtypedecl(p):
    "newtypedecl : NEWTYPE CONID EQUAL constructor"
    p[0] = [p[1], [p[2], [p[4]]]]


# Error rule for syntax errors
def p_error(p):
    print("Syntax error in input.")


parser = yacc.yacc(debug=0, write_tables=0)


# Type Declarations
# ------------------------------------------------------------------------------
def split(src):
    def keep(line):
        prefixes = [" ", "data ", "newtype ", "type "]
        return any(line.startswith(prefix) for prefix in prefixes)

    src = "\n".join(line for line in src.splitlines() if keep(line))
    type_decls = []
    for line in src.splitlines():
        if not line.startswith(" "):
            type_decls.append(line)
        else:
            type_decls[-1] = type_decls[-1] + "\n" + line
    return type_decls


def parse(src):
    if not isinstance(src, str):  # unicode in Python 2
        src = str(src)
    return [parser.parse(type_decl) for type_decl in split(src)]


def docstring(decl):
    if isinstance(decl, str):
        return decl
    else:
        assert isinstance(decl, list)
        if decl[0] == "data" or decl[0] == "newtype":
            type_name = decl[1][0]
            constructors = decl[1][1]
            _docstring = ""
            for i, constructor in enumerate(constructors):
                if i == 0:
                    prefix = type_name + " = "
                else:
                    prefix = " " * len(type_name) + " | "
                if i > 0:
                    _docstring += "\n"
                _docstring += prefix + docstring(constructor)
            return _docstring
        elif decl[0] == "type":
            return "{0} = {1}".format(decl[1][0], docstring(decl[1][1]))
        elif decl[0] == "list":
            return "[{0}]".format(docstring(decl[1][0]))
        elif decl[0] == "tuple":
            _types = [docstring(_type) for _type in decl[1]]
            _types = ", ".join(_types)
            return "({0})".format(_types)
        elif decl[0] == "map":
            # print(">>>", decl)
            key_type, value_type = decl[1]
            return "{{{0}: {1}}}".format(docstring(key_type), docstring(value_type))
        else:  # constructor, distinguish normal and record types
            type_name = decl[0]
            args_type = decl[1][0]
            args = decl[1][1]
            if args_type == "list":
                return "{0}({1})".format(
                    type_name, ", ".join(docstring(t) for t in args)
                )
            else:
                assert args_type == "map"
                args = [item for _, item in args]
                return "{0}({1})".format(
                    type_name, ", ".join(docstring(t) for t in args)
                )



# coding: utf-8

# Python 3 Standard Library
import builtins
import collections
import inspect
import pydoc
import sys

# Third-Party Libraries
import pkg_resources

# Pandoc
import pandoc
import pandoc.utils


# Haskell Type Constructs
# ------------------------------------------------------------------------------
def _fail_init(self, *args, **kwargs):
    type_name = type(self).__name__
    error = "cannot instantiate abstract type {type}"
    raise NotImplementedError(error.format(type=type_name))


class MetaType(type):
    def __repr__(cls):
        doc = getattr(cls, "__doc__", None)
        if doc is not None:
            return doc
        else:
            return type.__repr__(cls)


Type = MetaType("Type", (object,), {"__init__": _fail_init})


class Data(Type):
    pass


class Constructor(Data):
    def __init__(self, *args):
        if type(self) is Constructor:
            _fail_init(self, *args)
        else:
            self._args = list(args)

    def __iter__(self):
        return iter(self._args)

    def __getitem__(self, key):
        return self._args[key]

    def __setitem__(self, key, value):
        self._args[key] = value

    def __len__(self):
        return len(self._args)

    def __eq__(self, other):
        return type(self) == type(other) and self[:] == other[:]

    def __neq__(self, other):
        return not (self == other)

    def __repr__(self):
        typename = type(self).__name__
        args = ", ".join(repr(arg) for arg in self)
        return "{0}({1})".format(typename, args)

    __str__ = __repr__


class TypeDef(Type):
    pass


# Pandoc Types
# ------------------------------------------------------------------------------

_types_dict = {}


def _make_builtin_types():
    "Create Builtin Types"
    td = _types_dict
    td["Bool"] = bool
    td["Double"] = float
    td["Int"] = int
    try:
        td["String"] = unicode
        td["Text"] = unicode
    except NameError:
        td["String"] = str
        td["Text"] = str
    td["list"] = list
    td["tuple"] = tuple
    td["map"] = dict


def clear_types():
    """Uninstall Pandoc Types"""
    global _types_dict
    globs = globals()

    # Uninstall registered types
    for type_name in _types_dict:
        del globs[type_name]
    _types_dict = {}


def make_types():
    """Create Pandoc Types"""

    global _types_dict
    globs = globals()

    # Uninstall existing types (if any)
    clear_types()

    # Get the pandoc types version
    version = pandoc._configuration["pandoc_types_version"]

    # Create builtin types
    _make_builtin_types()

    # Load & parse the types definition
    defs_src = pandoc.utils.definitions[version]
    if not isinstance(defs_src, str):  # resource loaded as bytes in Python 3
        defs_src = defs_src.decode("utf-8")
    defs = pandoc.utils.parse(defs_src)

    # Create the types
    for decl in defs:
        decl_type = decl[0]
        type_name = decl[1][0]
        _dict = {"_def": decl, "__doc__": pandoc.utils.docstring(decl)}
        if decl_type in ("data", "newtype"):
            data_type = type(type_name, (Data,), _dict)
            _types_dict[type_name] = data_type
            # Remark: when there is a constructor with the same name as its
            #         data type, the data type is shadowed.
            #         This is intentional, but it's only consistent because
            #         it happens when there is a single constructor.
            # TODO: add an assert / check for this condition.
            for constructor in decl[1][1]:
                constructor_name = constructor[0]
                bases = (Constructor, data_type)
                _dict = {
                    "_def": constructor,
                    "__doc__": pandoc.utils.docstring(constructor),
                }
                type_ = type(constructor_name, bases, _dict)
                _types_dict[constructor_name] = type_
        elif decl_type == "type":
            type_ = type(type_name, (TypeDef,), _dict)
            _types_dict[type_name] = type_

    # Install the types
    globs.update(_types_dict)


# Create Types
# ------------------------------------------------------------------------------
_configuration = pandoc.configure(read=True)
if _configuration is None:
    pandoc.configure(auto=True)
else:
    pandoc.configure(**_configuration)



# coding: utf-8

# Python 3 Standard Library
import argparse
import collections
import copy
import inspect
import json
import os.path
import shlex
import shutil
import sys
import time
import tempfile

# Third-Party Libraries
import plumbum

# Pandoc
import pandoc.about
from . import utils

# TODO
# ------------------------------------------------------------------------------
# 
#   - Add pandoc options to the CLI (use shlex or partial parsing in argparse ?)
#

# Filesystem Helper
# ------------------------------------------------------------------------------
def rmtree(path):
    """Deal with Windows 
    (see e.g <https://www.gitmemory.com/issue/sdispater/poetry/1031/488759621>
    """
    retries = 10
    for i in range(retries - 1):
        try:
            shutil.rmtree(path)
            return
        except OSError:
            time.sleep(0.1)
    shutil.rmtree(path)


# Configuration
# ------------------------------------------------------------------------------
_configuration = None


def import_types():
    if configure(read=True) is None:
        configure(auto=True)
    import pandoc.types as types

    return types


def configure(
    auto=None,
    path=None,
    version=None,
    pandoc_types_version=None,
    read=False,
    reset=False,
):
    global _configuration

    default = (
        auto is None
        and path is None
        and version is None
        and pandoc_types_version is None
        and read is False
        and reset is False
    )
    if default:
        error = "configure expects at least one argument."
        raise ValueError(error)

    if reset is True:
        _configuration = None  # TODO: clean the types
        return

    read_only = (
        read
        and auto is None
        and path is None
        and version is None
        and pandoc_types_version is None
    )

    if auto:
        try:
            pandoc = plumbum.local["pandoc"]
            found_path = str(pandoc.executable)
        except plumbum.CommandNotFound as error:
            message = "cannot find the pandoc program.\n"
            paths = [str(p) for p in error.path]
            message += "paths:" + str(paths)
            raise RuntimeError(message)
        if path is None:
            path = found_path
        elif path != found_path:
            error = "found path {0!r} with auto=True "
            error += "but it doesn't match path={1!r}."
            raise ValueError(error.format(found_path, path))

    if path is not None:
        # TODO: manage invalid path
        pandoc = plumbum.machines.LocalCommand(path, "utf-8")
        found_version = pandoc("--version").splitlines()[0].split(" ")[1]
        if version is None:
            version = found_version
        elif version != found_version:
            error = "the version of the pandoc program is {0!r}"
            error += "but it doesn't match version={1!r}."
            raise ValueError(error.format(found_version, version))

    if version is not None:
        found_pandoc_types_versions = utils.resolve(version)
        if pandoc_types_version is None:
            if len(found_pandoc_types_versions) >= 1:
                # pick latest (ignore the real one that may be unknown)
                pandoc_types_version = found_pandoc_types_versions[-1]
            else:
                error = "cannot find a version of pandoc-types "
                error += "matching pandoc {0}"
                raise ValueError(error.format(version))
        elif pandoc_types_version not in found_pandoc_types_versions:
            error = "the version of pandoc is {0!r}"
            error += "but it doesn't match pandoc_types_version={1!r}."
            raise ValueError(error.format(version, pandoc_types_version))

    if not read_only:  # set the configuration, update pandoc.types

        try:
            from . import types
        except ImportError:  # only sensible explanation:
            # the types module is actually being imported (interpreted)
            # and is calling configure.
            types = sys.modules["pandoc.types"]

        _configuration = {
            "auto": auto,
            "path": path,
            "version": version,
            "pandoc_types_version": pandoc_types_version,
        }

        types.make_types()

    if read:
        return copy.copy(_configuration)


# JSON Reader / Writer
# ------------------------------------------------------------------------------

# Break compat? Read consumes markdown only? Bump major version number then.
# Yay, worth it.
#
# TODO: optional input or output FILES or FILENAMES in read/write? Dunno.
#       Think about it. The NAMES read and write seem to imply it ...
#       But the filesystem stuff is orthogonal really ...
#       However, for a ".doc" output for example, writing it as a string
#       is *probably* useless.
#
# TODO: Study also the str vs bytes stuff: we don't want encoding stuff
#       mixed in when we produce a word document, just BYTES.
#       For Markdown OTOH, unicode repr is the right abstraction.
#       What to do for latex, html docs? Bytes or Unicode?
#       FYI, Pandoc is using DECLARING utf-8 encoding for both in standalone
#       mode, so if you write these standalone outputs, you SHALL use utf-8 ...
#       Otherwise, well, I don't know ... but it's pretty much the same for
#       markdown: to get it properly processed, pandoc REQUIRES utf-8.
#       So, distinguish, markdown, latex and html as "source formats" and
#       use unicode for them? And bytes for the others?
#       What is the list? There is also ReST? How to get it automatically?
#       Try to trap the error? (Assuming the error message are stable?)
#       Nota: from the pandoc source code, ATM, "non-text formats" are
#        ["odt","docx","epub2","epub3","epub","pptx"]
#       But the non-text format categorization is used for OUTPUTS only,
#       what about inputs? OK, there is a classification into StringReader
#       (text sources) and ByteStringReader. For the latter, piping is not
#       accepted. So where is the list of the types of readers?
#       Grepping the sources leads to "docx", "odt" and "epub". OK then.
#       Am I really willing to hardcode all this stuff, or shall I return
#       bytes and let the user decide what to do with it? For INPUTS,
#       I can still accept unicode and convert to utf-8 seemlessly, the
#       question is: what to do for outputs? Only return unicode for markdown?
#       (that has no encoding metadata)? Dunno ...
#       UPDATE: OK, I have configured plumbum to always use utf-8 when
#       there is some conversion to be made between unicode and bytes.
#       BUT how can I deal with stuff (in or out) that are BYTES that
#       may not be utf-8 stuff?
#       Also, I forgot to configure cat for utf-8 ... and is cat available
#       on windows? Use temp files instead, that will solve two issues at
#       the same time (bytes as input and car availability).
#       Arf for the output, this is funny: for docx for example,
#       pandoc (haskell) WONT LET ME USE STDOUT! Which is nice :)
#       Nota: it won't read it either; so basically it manages differently
#       the binary formats. Same thing for epub for examples.
#       The messages are typically "pandoc: Cannot read archive from stdin"
#       and "Specify an output file using the -o option".
#       So I have to find a list in pandoc of binary vs text/utf-8 formats.
#       OR detect the appropriate error at runtime?
#       And then the bytes vs unicode policy is clear.
#       And I don't need to tweak encoding settings in plumbum since I
#       will use files for input and output anyway.
#
#       OK, so it's probaly safe to consider a shortlist of "binary" formats
#       that are "doc*", "epub*", "ppt*", "odt" and to return bytes only
#       for these formats.
#
#       And yes, working directly with filenames/files should work out
#       of the box, and yes, NOT using files should be ok too (and is
#       still my preferences: it should be simpler; if you need files,
#       use a proper keyword argument).

# TODO: add ".py" / Python support

_readers = {
    ".xhtml": "html",
    ".html": "html",
    ".htm": "html",
    ".md": "markdown",
    ".markdown": "markdown",
    ".muse": "muse",
    ".tex": "latex",
    ".latex": "latex",
    ".ltx": "latex",
    ".rst": "rst",
    ".org": "org",
    ".lhs": "markdown+lhs",
    ".db": "docbook",
    ".opml": "opml",
    ".wiki": "mediawiki",
    ".dokuwiki": "dokuwiki",
    ".textile": "textile",
    ".native": "native",
    ".json": "json",
    ".docx": "docx",
    ".t2t": "t2t",
    ".epub": "epub",
    ".odt": "odt",
    ".pdf": "pdf",
    ".doc": "doc",
}


def default_reader_name(filename):
    _, ext = os.path.splitext(filename)
    return _readers.get(ext)

def read(source=None, file=None, format=None, options=None):
    if configure(read=True) is None:
        configure(auto=True)
    if options is None:
        options = []

    filename = None
    if source is None:
        if file is None:
            raise ValueError("source or file should be defined.")
        if not hasattr(file, "read"):
            filename = file
            file = open(filename, "rb")
        source = file.read()
    else:
        if file is not None:
            raise ValueError("source or file should be defined, not both.")

    tmp_dir = tempfile.mkdtemp()
    if not isinstance(source, bytes):
        source = source.encode("utf-8")
    input_path = os.path.join(tmp_dir, "input")
    input = open(input_path, "wb")
    input.write(source)
    input.close()

    if format is None and filename is not None:
        format = default_reader_name(filename)
    if format is None:
        format = "markdown"
    if format != "json" and _configuration["path"] is None:
        error = "reading the {0!r} format requires the pandoc program"
        raise RuntimeError(error.format(format))

    if format == "json":
        json_file = open(input_path, "r", encoding="utf-8")
    else:
        if _configuration["path"] is None:
            error = "reading the {0!r} format requires the pandoc program"
            raise RuntimeError(error.format(format))
        pandoc = plumbum.machines.LocalCommand(_configuration["path"])
        output_path = os.path.join(tmp_dir, "output.js")
        options = (
            ["-t", "json", "-o", output_path]
            + list(options)
            + ["-f", format, input_path]
        )
        pandoc(options)
        json_file = open(output_path, "r", encoding="utf-8")
    json_ = json.load(json_file)
    json_file.close()
    rmtree(tmp_dir)
    if utils.version_key(_configuration["pandoc_types_version"]) < [1, 17]:
        return read_json_v1(json_)
    else:
        return read_json_v2(json_)


# TODO: add ".py" / Python support

_writers = {
    "": "markdown",
    ".pdf": "latex",
    ".tex": "latex",
    ".latex": "latex",
    ".ltx": "latex",
    ".context": "context",
    ".ctx": "context",
    ".rtf": "rtf",
    ".rst": "rst",
    ".s5": "s5",
    ".native": "native",
    ".json": "json",
    ".txt": "markdown",
    ".text": "markdown",
    ".md": "markdown",
    ".muse": "muse",
    ".markdown": "markdown",
    ".textile": "textile",
    ".lhs": "markdown+lhs",
    ".texi": "texinfo",
    ".texinfo": "texinfo",
    ".db": "docbook",
    ".odt": "odt",
    ".docx": "docx",
    ".epub": "epub",
    ".org": "org",
    ".asciidoc": "asciidoc",
    ".adoc": "asciidoc",
    ".fb2": "fb2",
    ".opml": "opml",
    ".icml": "icml",
    ".tei.xml": "tei",
    ".tei": "tei",
    ".ms": "ms",
    ".roff": "ms",
    ".pptx": "pptx",
    ".xhtml": "html",
    ".html": "html",
    ".htm": "html",
}


def default_writer_name(filename):
    if filename.endswith(".tei.xml"):
        filename = filename[:-4]  # uhu ? test this.
    _, ext = os.path.splitext(filename)
    if len(ext) == 2 and ext[1] in "0123456789":
        return "man"
    else:
        return _writers.get(ext)


# TODO: better management for pdf "format" which is not a format according
#       to pandoc ... ("latex" or "beamer" are, pdf is hidden in the filename
#       extension)


def write(doc, file=None, format=None, options=None):
    if options is None:
        options = []

    types = import_types()

    if isinstance(doc, types.Inline):
        inline = doc
        doc = types.Plain([inline])
    if isinstance(doc, types.Block):
        block = doc
        doc = types.Pandoc(types.Meta({}), [block])
    if not isinstance(doc, types.Pandoc):
        # TODO: support lists of inlines and blocks ?
        raise TypeError(f"{doc!r} is not a Pandoc, Block or Inline instance.")

    tmp_dir = tempfile.mkdtemp()
    filename = None
    if file is not None and not hasattr(file, "write"):
        filename = file
        file = open(filename, "wb")

    if format is None and filename is not None:
        format = default_writer_name(filename)
    if format is None:
        format = "markdown"  # instead of html, yep.
    if format != "json" and _configuration["path"] is None:
        error = "writing the {0!r} format requires the pandoc program"

    configuration = configure(read=True)
    if utils.version_key(configuration["pandoc_types_version"]) < [1, 17]:
        json_ = write_json_v1(doc)
    else:
        json_ = write_json_v2(doc)
    json_str = json.dumps(json_)
    input_path = os.path.join(tmp_dir, "input.js")
    input = open(input_path, "wb")
    input.write(json_str.encode("utf-8"))
    input.close()

    if format == "json":
        output_path = input_path
    else:
        if filename is not None:
            # preserve extensions (sometimes pandoc looks for the extension,
            # e.g. for pdf files)
            tmp_filename = os.path.basename(filename)
        else:
            tmp_filename = "output"
        pandoc = plumbum.machines.LocalCommand(_configuration["path"])
        output_path = os.path.join(tmp_dir, tmp_filename)
        options = (
            ["-t", format, "-o", output_path]
            + list(options)
            + ["-f", "json", input_path]
        )
        pandoc(options)

    output_bytes = open(output_path, "rb").read()
    binary_formats = ["doc", "epub", "ppt", "odt"]
    if any(tag in format for tag in binary_formats) or output_path.endswith(".pdf"):
        output = output_bytes
    else:  # text format
        output = output_bytes.decode("utf-8")
    rmtree(tmp_dir)

    if file is not None:
        file.write(output_bytes)
    return output


# JSON Reader v1
# ------------------------------------------------------------------------------
def read_json_v1(json_, type_=None):
    types = import_types

    if type_ is None:
        type_ = types.Pandoc
    if isinstance(type_, str):
        type_ = getattr(types, type_)
    if not isinstance(type_, list):  # not a type def (yet).
        if issubclass(type_, types.Type):
            type_ = type_._def
        else:  # primitive type
            return type_(json_)

    if type_[0] == "type":  # type alias
        type_ = type_[1][1]
        return read_json_v1(json_, type_)
    if type_[0] == "list":
        item_type = type_[1][0]
        return [read_json_v1(item, item_type) for item in json_]
    if type_[0] == "tuple":
        tuple_types = type_[1]
        return tuple(
            read_json_v1(item, item_type)
            for (item, item_type) in zip(json_, tuple_types)
        )
    if type_[0] == "map":
        key_type, value_type = type_[1]
        return types.map(
            [
                (read_json_v1(k, key_type), read_json_v1(v, value_type))
                for (k, v) in json_.items()
            ]
        )

    data_type = None
    constructor = None
    if type_[0] in ("data", "newtype"):
        data_type = type_
        constructors = data_type[1][1]
        if len(constructors) == 1:
            constructor = constructors[0]
        else:
            constructor = getattr(types, json_["t"])._def
    elif type_[0][0] == type_[0][0].upper():
        constructor = type_
        constructor_type = getattr(types, constructor[0])
        data_type = constructor_type.__mro__[2]._def

    single_type_constructor = len(data_type[1][1]) == 1
    single_constructor_argument = len(constructor[1][1]) == 1
    is_record = constructor[1][0] == "map"

    json_args = None
    args = None
    if not is_record:
        if single_type_constructor:
            json_args = json_
        else:
            json_args = json_["c"]
        if single_constructor_argument:
            json_args = [json_args]
        args = [read_json_v1(jarg, t) for jarg, t in zip(json_args, constructor[1][1])]
    else:
        keys = [k for k, t in constructor[1][1]]
        types_ = [t for k, t in constructor[1][1]]
        json_args = [json_[k] for k in keys]
        args = [read_json_v1(jarg, t) for jarg, t in zip(json_args, types_)]
    C = getattr(types, constructor[0])
    return C(*args)


# JSON Writer v1
# ------------------------------------------------------------------------------
def write_json_v1(object_):
    types = import_types()

    odict = collections.OrderedDict
    type_ = type(object_)
    if not isinstance(object_, types.Type):
        if isinstance(object_, (list, tuple)):
            json_ = [write_json_v1(item) for item in object_]
        elif isinstance(object_, dict):
            json_ = odict((k, write_json_v1(v)) for k, v in object_.items())
        else:  # primitive type
            json_ = object_
    else:
        constructor = type(object_)._def
        data_type = type(object_).__mro__[2]._def
        single_type_constructor = len(data_type[1][1]) == 1
        single_constructor_argument = len(constructor[1][1]) == 1
        is_record = constructor[1][0] == "map"

        json_ = odict()
        if not single_type_constructor:
            json_["t"] = type(object_).__name__

        if not is_record:
            c = [write_json_v1(arg) for arg in object_]
            if single_constructor_argument:
                c = c[0]
            if single_type_constructor:
                json_ = c
            else:
                json_["c"] = c
        else:
            keys = [kt[0] for kt in constructor[1][1]]
            for key, arg in zip(keys, object_):
                json_[key] = write_json_v1(arg)
    return json_


# JSON Reader v2
# ------------------------------------------------------------------------------
def read_json_v2(json_, type_=None):
    types = import_types()

    if type_ is None:
        type_ = types.Pandoc
    if isinstance(type_, str):
        type_ = getattr(types, type_)
    if not isinstance(type_, list):  # not a type def (yet).
        if issubclass(type_, types.Type):
            type_ = type_._def
        else:  # primitive type
            return type_(json_)

    if type_[0] == "type":  # type alias
        type_ = type_[1][1]
        return read_json_v2(json_, type_)
    if type_[0] == "list":
        item_type = type_[1][0]
        return [read_json_v2(item, item_type) for item in json_]
    if type_[0] == "tuple":
        tuple_types = type_[1]
        return tuple(
            read_json_v2(item, item_type)
            for (item, item_type) in zip(json_, tuple_types)
        )
    if type_[0] == "map":
        key_type, value_type = type_[1]
        return types.map(
            [
                (read_json_v2(k, key_type), read_json_v2(v, value_type))
                for (k, v) in json_.items()
            ]
        )

    data_type = None
    constructor = None
    if type_[0] in ("data", "newtype"):
        data_type = type_
        constructors = data_type[1][1]
        if len(constructors) == 1:
            constructor = constructors[0]
        else:
            constructor = getattr(types, json_["t"])._def
    elif type_[0][0] == type_[0][0].upper():
        constructor = type_
        constructor_type = getattr(types, constructor[0])
        data_type = constructor_type.__mro__[2]._def

    single_type_constructor = len(data_type[1][1]) == 1
    single_constructor_argument = len(constructor[1][1]) == 1
    is_record = constructor[1][0] == "map"

    json_args = None
    args = None
    if constructor[0] == "Pandoc":
        # TODO; check API version compat
        meta = read_json_v2(json_["meta"], types.Meta)
        blocks = read_json_v2(json_["blocks"], ["list", ["Block"]])
        return types.Pandoc(meta, blocks)
    elif constructor[0] == "Meta":
        type_ = ["map", ["String", "MetaValue"]]
        return types.Meta(read_json_v2(json_, type_))
    elif not is_record:
        if single_type_constructor:
            json_args = json_
        else:
            json_args = json_.get("c", [])
        if single_constructor_argument:
            json_args = [json_args]
        args = [read_json_v2(jarg, t) for jarg, t in zip(json_args, constructor[1][1])]
    else:
        keys = [k for k, t in constructor[1][1]]
        types_ = [t for k, t in constructor[1][1]]
        json_args = [json_[k] for k in keys]
        args = [read_json_v2(jarg, t) for jarg, t in zip(json_args, types_)]
    C = getattr(types, constructor[0])
    return C(*args)


# JSON Writer v2
# ------------------------------------------------------------------------------
def write_json_v2(object_):
    types = import_types()

    odict = collections.OrderedDict
    type_ = type(object_)
    if not isinstance(object_, types.Type):
        if isinstance(object_, (list, tuple)):
            json_ = [write_json_v2(item) for item in object_]
        elif isinstance(object_, dict):
            json_ = odict((k, write_json_v2(v)) for k, v in object_.items())
        else:  # primitive type
            json_ = object_
    elif isinstance(object_, types.Pandoc):
        version = configure(read=True)["pandoc_types_version"]
        metadata = object_[0]
        blocks = object_[1]
        json_ = odict()
        json_["pandoc-api-version"] = [int(n) for n in version.split(".")]
        json_["meta"] = write_json_v2(object_[0][0])
        json_["blocks"] = write_json_v2(object_[1])
    else:
        constructor = type(object_)._def
        data_type = type(object_).__mro__[2]._def
        single_type_constructor = len(data_type[1][1]) == 1
        has_constructor_arguments = len(constructor[1][1]) >= 1
        single_constructor_argument = len(constructor[1][1]) == 1
        is_record = constructor[1][0] == "map"

        json_ = odict()
        if not single_type_constructor:
            json_["t"] = type(object_).__name__
        if not is_record:
            c = [write_json_v2(arg) for arg in object_]
            if single_constructor_argument:
                c = c[0]
            if single_type_constructor:
                json_ = c
            else:
                if has_constructor_arguments:
                    json_["c"] = c
        else:
            keys = [kt[0] for kt in constructor[1][1]]
            for key, arg in zip(keys, object_):
                json_[key] = write_json_v2(arg)
    return json_


# Iteration
# ------------------------------------------------------------------------------

# Thoughts:
#
#   - maybe 'iter_path' is not the right naming:
#     I would envision for iter_path something that returns roughly speaking
#     a list of indices that would allow us to find the element from doc.
#     What would be the right new name for the old concept of path then?
#     Have the list of ancestors named "context"? for example?
#     And have it as supplementary data (also return elt as usual?)
#     So consider a "mega-iter-named-whataver-with-options" ?
#     And keep "iter" for the building block? Or use 'iter' for that?
#     Wrt path as list of indices: have [] (or another function) support
#     these list of path directly? Yeah would need some support anyway.
#
#   - Nota: I have at least 3 different "context" structures that may
#     be handy:
#
#       - the list of parents (say without the elt, aka "context" or "parents"?)
#
#       - the list of indices (aka "path")
#
#       - a hybrid parent + index tuple list.
#
#     And these lists *may* (or may not) be reversed for convenience.
#     Actually we may keep the top to bottom order and name "path"
#     the list of pairs. This is a redundant structure (all you need
#     for example is the root and the indices), but we don't care,
#     the redundancy is actually convenient.
#
#     And add a "path" option to iter to return the (elt, path) pair.
#     Does it fly? Name ? "path" or "context"? Nah, path ...
#
#   - Nota: does it make sense to also use string keys in path for maps?
#     Several issues: have a look in the complete chain to see if pandoc
#     map-like structure have their ordered preserved (I think not),
#     then think "tree/structural" vs value stuff (maps have a varying
#     number of items for example ...), and finally think of the consistency
#     of the interface of named keys wrt to "iter" that is 100% integer based.
#     Nota wrt "non-structural iteration": we are already doing it with lists,
#     so nothing new here.
#
#  - TODO: explore visitor-style stuff (have a look at ANTLR for example)
#
#  - TODO: explore functional programming style (e.g. fmap for trees, etc.)
#


def iter(elt, path=False, enter=None, exit=None):
    if path is not False:
        if not isinstance(path, list):  # e.g. path = True
            path = []

    args = [elt]
    if path is not False:
        args.append(path)

    if enter is not None:
        enter(*args)

    if path is False:
        yield elt
    else:
        yield elt, path

    if isinstance(elt, dict):
        elt = elt.items()
    if hasattr(elt, "__iter__") and not isinstance(elt, types.String):
        for i, child in enumerate(elt):
            if path is False:
                child_path = False
            else:
                child_path = path.copy() + [(elt, i)]
            for subelt in iter(child, path=child_path, enter=enter, exit=exit):
                yield subelt

    if exit is not None:
        exit(*args)


def iter_path(elt):
    path = []

    def enter(elt_):
        path.append(elt_)

    def exit(elt_):
        path.pop()

    for elt_ in iter(elt, enter=enter, exit=exit):
        yield path


def get_parent(doc, elt):
    for path in iter_path(doc):
        elt_ = path[-1]
        if elt is elt_:
            parent = path[-2] if len(path) >= 2 else None
            return parent


# Main Entry Point
# ------------------------------------------------------------------------------
# TODO : support custom indendation for Python output.
# BUG : by using files instead of filenames in argparse types,
#       we are losing the extension info and cannot select the proper format.
#       This is an issue in itself ; but since we use the pandoc convention
#       not to have "pdf" as an output format per se, that means that we
#       cannot output pdf at all.
def main():
    prog = "python -m pandoc"
    description = "Read/write pandoc documents with Python"
    parser = argparse.ArgumentParser(prog=prog, description=description)
    parser.set_defaults(command=None)
    subparsers = parser.add_subparsers()
    read_parser = subparsers.add_parser("read")
    read_parser.set_defaults(command="read")
    read_parser.add_argument(
        "file", nargs="?", metavar="FILE", default=None, help="input file",
    )
    read_parser.add_argument(
        "-f", "--format", nargs="?", default=None, help="input format",
    )
    read_parser.add_argument(
        "-o", "--output", nargs="?", default=None, help="output file",
    )
    write_parser = subparsers.add_parser("write")
    write_parser.set_defaults(command="write")
    write_parser.add_argument(
        "file", nargs="?", metavar="FILE", default=None, help="input file",
    )
    write_parser.add_argument(
        "-f", "--format", nargs="?", default=None, help="output format",
    )
    write_parser.add_argument(
        "-o", "--output", nargs="?", default=None, help="output file",
    )
    args = parser.parse_args()
    if args.command == "read":
        if args.file is None:
            file = sys.stdin
        else:
            file = args.file
        doc = read(file=file, format=args.format)
        content = str(doc) + "\n"
        if args.output is None:
            output = sys.stdout.buffer
        else:
            output = open(args.output, "wb")
        assert "b" in output.mode
        content = content.encode("utf-8")
        output.write(content)
    elif args.command == "write":
        if args.file is None:
            # We always interpret the standard input stream as utf-8 ;
            # see <https://pandoc.org/MANUAL.html#character-encoding>
            file = sys.stdin.buffer  # sys.stdin may not be utf-8.
            assert "b" in file.mode
            doc_bytes = file.read()
            doc_string = doc_bytes.decode("utf-8")
        else:
            file = open(args.file, mode="r", encoding="utf-8")
            doc_string = file.read()
        types = import_types()
        globs = types.__dict__.copy()
        doc = eval(doc_string, globs)
        if args.output is None:
            output = sys.stdout.buffer
        else:
            output = args.output
        write(doc, file=output, format=args.format)



#!/usr/bin/env python
# coding: utf-8

# Pip Package Manager
# ------------------------------------------------------------------------------
try:
     import pip
     import setuptools
except ImportError:
    url = "http://pip.readthedocs.org"
    error = f"pip is not installed, refer to <{url}> for instructions."
    raise ImportError(error)

# Pandoc Metadata
# ------------------------------------------------------------------------------
m = {}
exec(open("src/pandoc/about.py").read(), m)
metadata = dict((k, v) for (k, v) in m.items() if not k.startswith('_'))

# Setup Configuration
# ------------------------------------------------------------------------------
contents = {
  "packages": setuptools.find_packages("src"),
  "package_dir": {"": "src"},
  "package_data": {"pandoc": ["pandoc-types.js", "tests.md"]},
}
requirements = {
  "install_requires": ["plumbum", "ply"],
}

info = {}
info.update(metadata)
info.update(contents)
info.update(requirements)

info["long_description"] = open("README.md", encoding="utf-8").read()
info["long_description_content_type"] = 'text/markdown'

if __name__ == "__main__":
    setuptools.setup(**info)


# coding: utf-8

name = "pandoc"
version = "2.0a4"
license = "MIT License"
description = "Pandoc Documents for Python"
author = "Sébastien Boisgérault"
author_email = "Sebastien.Boisgerault@mines-paristech.fr"
classifiers = [
    "Programming Language :: Python :: 3",
    "Development Status :: 3 - Alpha",
    "Topic :: Software Development",
    "Topic :: Text Editors :: Text Processing",
]




# Python 2.7 Directives
from __future__ import absolute_import

# Python 2.7 Standard Library
import ast
import doctest
import json
from subprocess import Popen, PIPE
import sys

# Local Library
import pandoc
from pandoc.types import *

# This doctest extension require pandoc 1.16
from subprocess import Popen, PIPE
p = Popen(["pandoc", "-v"], stdout=PIPE)
if b"pandoc 1.16" not in p.communicate()[0]:
    raise RuntimeError("pandoc 1.16 not found")

# TODO: 
#    - optional round-trip check (global setting)
#    - "inject" pandoc.read errors in the doctest so that we can catch them.
#      Mmmm we probably can't do it, unless we do some fugly stack patching.
#      So what ? pandoc.read on any output JSON generated by pandoc is not 
#      supposed to fail ... Implement a "safe" read on top of it ?
#      Return a reference error message ? (maybe the full traceback?)
#      And make sure that after that, it is properly handled?

# ------------------------------------------------------------------------------

# Global Variables
check_round_trip = True

# New Doctest Directive: PANDOC 
PANDOC = doctest.register_optionflag("PANDOC")
doctest.PANDOC = PANDOC
doctest.__all__.append("PANDOC")
doctest.COMPARISON_FLAGS = doctest.COMPARISON_FLAGS | PANDOC

# Helpers
from subprocess import Popen, PIPE
import json
def to_json(txt):
    p = Popen(["pandoc", "-tjson"], 
              stdout=PIPE, stdin=PIPE, stderr=PIPE)
    json_string = p.communicate(input=txt.encode("utf-8"))[0].decode("utf-8")
    json_doc = json.loads(json_string)
    return json_doc

def linebreak(text, length=80):
    text = text.replace(u"\n", "")
    chunks = [text[i:i+length] for i in range(0, len(text), length)]
    return "\n".join(chunks) + "\n"

# Implement the pandoc output checker and monkey-patch doctest:
_doctest_OutputChecker = doctest.OutputChecker
class PandocOutputChecker(_doctest_OutputChecker):

    def round_trip_check(self, json_doc):
        json_doc_2 = None
        try:
            doc = pandoc.read(json_doc)
            json_doc_2 = pandoc.write(doc)
        except Exception as error:
            jsond_doc_2 = error
        if json_doc == json_doc_2 :
            return True
        else:
            return json_doc, json_doc_2

    def text_repr_to_docs(self, text_repr):
        text = ast.literal_eval(text_repr)
        if not isinstance(text, str):
            raise TypeError("{0!r} is not a string".format(text))
        json_doc = to_json(text)
        doc = pandoc.read(json_doc)
        return doc, json_doc

    def str_error(self, error):
        message = ""
        if len(error.args):
            message = error.args[0]
        return "Traceback (most recent call last)\n" + \
               "    ...\n"                           + \
               "{0}: {1}\n".format(type(error).__name__, message) 

    # TODO: need an eval-based checker for Python2/3 compatibility.
    def check_output(self, want, got, optionflags):
        if optionflags & PANDOC:
            try:
                want = eval(want.replace("\n", ""))
                doc, json_doc = self.text_repr_to_docs(got)
                got  = doc
                return got == want
            except Exception as error:
                got = self.str_error(error)
                return False
        super_check_output = _doctest_OutputChecker.check_output
        check = super_check_output(self, want, got, optionflags)
        if optionflags & PANDOC & check_round_trip:
            check = check and (self.round_trip_check(json_got) is True)
        return check

    def output_difference(self, example, got, optionflags):
        if optionflags & PANDOC:
            example.want = linebreak(example.want, 76)
            doc, json_doc = None, None
            try:
                doc, json_doc = self.text_repr_to_docs(got)
                got  = linebreak(repr(doc), 76)
                if check_round_trip:
                    check = self.round_trip_check(json_doc)
                    if check is not True:
                        json_doc, json_doc_2 = check
                        json_doc = linebreak(repr(json_doc), 72).strip()
                        if isinstance(json_doc_2, Exception):
                            json_doc_2 = self.str_error(json_doc_2)
                            json_doc_2 = "\n".join([4*" " + line for line in json_doc2.split("\n")])
                        else:
                            json_doc_2 = linebreak(repr(json_doc_2), 72).strip()
                        got = ("Traceback (most recent call last)\n" +\
                               "    ....\n"                          +\
                               "ValueError: failed JSON read+write round-trip:\n" +\
                               "    {0}\n" +\
                               "vs:\n    {1}\n").format(json_doc, json_doc_2)
            except Exception as error:
                got = self.str_error(error)
        super_output_difference = _doctest_OutputChecker.output_difference
        return super_output_difference(self, example, got, optionflags)

doctest.OutputChecker = PandocOutputChecker

